#include "stdafx.h"
#include <ablib/lexical_cast.h>
#include <algorithm>
#include "OLE.h"

extern OLE *g_OLE;
extern ablib::string tmp_string;

OLE::OLE(){
	::CoInitialize(NULL);
}

OLE::~OLE(){
	Clear();
	::CoUninitialize();
}

void OLE::Clear(){
	m_handles.clear();
}

OLECHAR *OLE::ole_string::GetOleChar(const TCHAR *str){
	if(str == NULL)return NULL;
	auto len = ::lstrlen(str);
	OLECHAR *buf;
#ifdef UNICODE
	buf = ::SysAllocStringLen(str,len + 1);
#else
	auto size = ::MultiByteToWideChar(CP_ACP,0,str,len,NULL,0);
	buf = ::SysAllocStringLen(NULL,size);
	::MultiByteToWideChar(CP_ACP,0,str,len,buf,size);
#endif
	return buf;
}

OLE::ole_string::ole_string(const TCHAR *str){
	m_buf = GetOleChar(str);
}

intptr_t OLE::Initialize(const TCHAR *obj){
	CLSID cls;
	ole_string buf(obj);
	HRESULT hr = ::CLSIDFromProgID(static_cast<OLECHAR*>(buf),&cls);
	if(FAILED(hr)){
		hr = ::CLSIDFromString(static_cast<OLECHAR*>(buf),&cls);
	}
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("͒mȂOLET[ołD"));
		return NULL;
	}
	
	IDispatch *disp = NULL;
	hr = ::CoCreateInstance(cls,NULL,CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER,
	IID_IDispatch,reinterpret_cast<void**>(&disp));
	//	hr = ::CoCreateInstance(cls,NULL,CLSCTX_ALL,IID_IDispatch,(void**)&disp);
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("̏ɎsD"));
		return NULL;
	}
	return m_handles.add(disp);
}

intptr_t OLE::Connect(const TCHAR *obj){
	CLSID cls;
	ole_string buf(obj);
	HRESULT hr = ::CLSIDFromProgID(static_cast<const OLECHAR*>(buf),&cls);
	if(FAILED(hr)){
		hr = ::CLSIDFromString(static_cast<const OLECHAR*>(buf),&cls);
	}
	if(FAILED(hr))return Bind(obj);
	
	IUnknown *unknown;
	hr = ::GetActiveObject(cls,0,&unknown);
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("͓Ă܂D"));
		return NULL;
	}
	IDispatch *disp = NULL;
	hr = unknown->QueryInterface(IID_IDispatch,reinterpret_cast<void**>(&disp));
	unknown->Release();
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("Ƃ̐ڑɎsD"));
		return NULL;
	}
	return m_handles.add(disp);
}

intptr_t OLE::Bind(const TCHAR *obj){
	IBindCtx *bindctx;
	HRESULT hr = ::CreateBindCtx(0,&bindctx);
	if(FAILED(hr)){
		SetmacrodllLastError(_T("CreateBindCtxɎs"));
		return NULL;
	}
	
	ULONG eaten;
	IMoniker *moniker;
	ole_string buf(obj);
	hr = ::MkParseDisplayName(bindctx,static_cast<OLECHAR*>(buf),&eaten,&moniker);

	if(FAILED(hr)){
		bindctx->Release();
		SetmacrodllLastError(ablib::string(obj) + _T("ɑ΂CMkParseDisplayNameŎsD"));
		return NULL;
	}
	IDispatch *disp;
	hr = moniker->BindToObject(bindctx,NULL,IID_IDispatch,reinterpret_cast<void**>(&disp));
	moniker->Release();
	bindctx->Release();
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("bindsD"));
		return NULL;
	}
	return m_handles.add(disp);
}


intptr_t OLE::SetPropertyStr(intptr_t handle,const TCHAR *prop,const TCHAR *val){
	variant var(val);
	HRESULT hr = Invoke(handle,prop,1,var.GetPointer(),NULL,DISPATCH_PROPERTYPUT);
	return FAILED(hr) ? FALSE : TRUE;
}

intptr_t OLE::SetPropertyInt(intptr_t handle,const TCHAR *prop, intptr_t val){
	variant var(val);
	HRESULT hr = Invoke(handle,prop,1,var.GetPointer(),NULL,DISPATCH_PROPERTYPUT);
	return FAILED(hr) ? FALSE : TRUE;
}


HRESULT OLE::Invoke(intptr_t handle, const TCHAR *func, intptr_t argc, VARIANTARG *argv, VARIANT *result, USHORT invoke_type){
	if(handle == 0){
		SetmacrodllLastError(_T("IuWFNgłD"));
		return FALSE;
	}
	ablib::string err;
	if(invoke_type == DISPATCH_METHOD)err = ablib::string(func) + _T("̎sɎs");
	else if(invoke_type == DISPATCH_PROPERTYGET)err = ablib::string(_T("vpeB")) + func + _T("擾ł܂ł");
	else if(invoke_type == DISPATCH_PROPERTYPUT)err = ablib::string(_T("vpeB")) + func + _T("ݒł܂ł");
	IDispatch *disp = m_handles[(int)handle];
	DISPID dispid;
	ole_string buf(func);
	OLECHAR *oc = static_cast<OLECHAR*>(buf);
	HRESULT hr = disp->GetIDsOfNames(IID_NULL,&oc,1,LOCALE_USER_DEFAULT,&dispid);
	if(FAILED(hr)){
		SetmacrodllLastError(err + _T("iGetIDsOfNamesŎsjD"));
		return 0;
	}
	
	DISPPARAMS dispParams;
	dispParams.rgvarg = argv;
	dispParams.rgdispidNamedArgs = NULL;
	dispParams.cArgs = (int)argc;
	dispParams.cNamedArgs = 0;
	
	DISPID dispIDNamedArgs[1] = {DISPID_PROPERTYPUT};
	if(invoke_type == DISPATCH_PROPERTYPUT){
		dispParams.cNamedArgs = 1;
		dispParams.rgdispidNamedArgs = dispIDNamedArgs;
	}
	
	EXCEPINFO excepinfo;
	::ZeroMemory(&excepinfo,sizeof(EXCEPINFO));
	UINT argerr = 0;
	hr = disp->Invoke(dispid,IID_NULL,LOCALE_USER_DEFAULT,invoke_type,&dispParams,result,&excepinfo,&argerr);
	if(FAILED(hr)){
		SetmacrodllLastError(err + _T("iInvokeŎsjD\n" + GetErrMsg(&excepinfo)));
		return 0;
	}
	return TRUE;
}

ablib::string OLE::GetErrMsg(EXCEPINFO *ei){
	if(ei->pfnDeferredFillIn)(ei->pfnDeferredFillIn)(ei);
	ablib::string rv;
	if(ei->bstrSource){
		rv += BSTR2String(ei->bstrSource);
	}else{
		rv += "ssource";
	}
	rv += "ɂāCR[h";
	if(ei->wCode == 0)rv += ablib::lexical_cast<ablib::string>(ei->scode);
	else rv += ablib::lexical_cast<ablib::string>(ei->wCode);
	rv += "ŃG[D";
	if(ei->bstrDescription)rv += "\n" + BSTR2String(ei->bstrDescription);
	return rv;
}


ablib::string OLE::BSTR2String(BSTR bstr){
#ifdef UNICODE
	return ablib::string(bstr);
#else
	ablib::string r;
	int len = ::lstrlenW(bstr);
	int size = ::WideCharToMultiByte(CP_ACP,0,bstr,len,NULL,0,NULL,NULL);
	char *s = new char[size + 1];
	::ZeroMemory(s,sizeof(char)*(size + 1));
	::WideCharToMultiByte(CP_ACP,0,bstr,len,s,size,NULL,NULL);
	r = s;
	delete[] s;
	return r.c_str();
#endif
}

intptr_t OLE::GetLastResultInt(){return static_cast<intptr_t>(m_result.var);}
intptr_t OLE::GetLastResultDisp(){return m_result.handle;}
const TCHAR *OLE::GetLastResultStr(){
	tmp_string = m_result.var;
	return tmp_string.c_str();
}

intptr_t OLE::GetLastResultBool(){return m_result.var->boolVal;}

#define GETARGS \
		argc = ::lstrlen(arglist);\
		if(argc > 0){\
			argv = new VARIANTARG[argc];\
			va_list vlists;\
			va_start(vlists,arglist);\
			for(int i = 0 ; i < argc ; ++i){\
				int j = argc - i - 1;\
				::VariantInit(argv + j);\
				switch(arglist[i]){\
				case 's':\
					argv[j].vt = VT_BSTR;\
					argv[j].bstrVal = OLE::ole_string::GetOleChar(va_arg(vlists,PTCHAR));\
					break;\
				case 'd':\
					argv[j].vt = VT_DISPATCH;\
					argv[j].pdispVal = g_OLE->GetDispatch(va_arg(vlists,int));\
					break;\
				case 'i':\
				default:\
					argv[j].vt = VT_I4;\
					argv[j].intVal = va_arg(vlists,int);\
					break;\
				}\
			}\
		}\
	
#define RELEASEARGS \
	for(int i = 0 ; i < argc ; ++i){\
		if(argv[i].vt == VT_BSTR)::SysFreeString(argv[i].bstrVal);\
		::VariantClear(argv + i);\
	}\
	delete[] argv;\


EXPORT intptr_t OLE_INVOKE(intptr_t handle,const TCHAR *func,const TCHAR *arglist,...){
	DEBUGSTRING(_T("macrodll.dll:OLE_INVOKE: handle = %d, func = `%s'"),handle,func);
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;

	if(arglist){
		GETARGS
	}
	
	HRESULT rv = g_OLE->Invoke_Sub(handle,func,argc,argv);

	RELEASEARGS
	
	return rv;
}

EXPORT intptr_t OLE_GETPROPERTYDISP(intptr_t handle,const TCHAR *prop,const TCHAR *arglist,...){
	DEBUGSTRING(_T("macrodll.dll:OLE_GETPROPERTYDISP: handle = %d, prop = `%s'..."),handle,prop);
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;
	if(arglist){
		GETARGS
	}
	OLE::variant var;
	HRESULT hr = g_OLE->GetProperty(handle,prop,var.GetPointer(),argc,argv);
	if(var->vt != VT_DISPATCH){
		SetmacrodllLastError(ablib::string(_T("vpeB")) + prop + _T("̌^VT_IDISPATCHł͂܂łD"));
	}
	IDispatch *rv = ((!FAILED(hr) && var->vt == VT_DISPATCH) ? var->pdispVal : NULL);
	RELEASEARGS
	if(rv == NULL)return 0;
	else return g_OLE->add_to_handles(rv);
}

EXPORT intptr_t OLE_GETPROPERTYINT(intptr_t handle,const TCHAR *prop,const TCHAR *arglist,...){
	DEBUGSTRING(_T("macrodll.dll:OLE_GETPROPERTYINT: handle = %d, prop = `%s'..."),handle,prop);
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;
	if(arglist){
		GETARGS
	}
	OLE::variant var;
	HRESULT hr = g_OLE->GetProperty(handle,prop,var.GetPointer(),argc,argv);
	int rv = var;
	RELEASEARGS
	return rv;
}

EXPORT const TCHAR *OLE_GETPROPERTYSTR(intptr_t handle, const TCHAR *prop,const TCHAR *arglist,...){
	DEBUGSTRING(_T("macrodll.dll:OLE_GETPROPERTYSTR: handle = %d, prop = `%s'..."),handle,prop);
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;
	if(arglist){
		GETARGS
	}
	OLE::variant var;
	HRESULT hr = g_OLE->GetProperty(handle,prop,var.GetPointer(),argc,argv);
	if(!FAILED(hr))tmp_string = var;
	else tmp_string = _T("");
	RELEASEARGS
	return tmp_string.c_str();
}

HRESULT OLE::GetProperty(intptr_t handle, const TCHAR *prop, VARIANT *var, intptr_t argc, VARIANTARG *argv){
	HRESULT rv = Invoke(handle,prop,argc,argv,var,DISPATCH_PROPERTYGET);
	if(!FAILED(rv) && var->vt == VT_DISPATCH){
		var->pdispVal->AddRef();
		m_handles.add(var->pdispVal);
	}
	return rv;
}

ablib::string OLE::variant::to_str() const{
	ablib::string rv;
	if(m_var.vt == VT_BSTR){
		rv = OLE::BSTR2String(m_var.bstrVal);
	}else if(m_var.vt == VT_DATE){
		DATE date;
		if(V_ISBYREF(&m_var))date = *V_DATEREF(&m_var);
		else date = V_DATE(&m_var);
		int y, m, d, hh, mm, ss;
		char szTime[20];
		OLE::double2time(date, &y, &m, &d, &hh, &mm, &ss);
		sprintf(szTime,"%4.4d/%02.2d/%02.2d %02.2d:%02.2d:%02.2d",y, m, d, hh, mm, ss);
		rv = szTime;
	}else{
		variant v;
		HRESULT hr = ::VariantChangeTypeEx(v.GetPointer(),&m_var,LOCALE_USER_DEFAULT,0,VT_BSTR);
		if(!FAILED(hr))rv = v;
		else{
			::SetmacrodllLastError(_T("ʂ̕ւ̕ϊɎsD"));
		}
	}
	return rv;
}

// win32olẽRsy
void OLE::jd2civil(long day, int *yy, int *mm, int *dd){
    long x, a, b, c, d, e;
    x = (long)(((double)day - 1867216.25) / 36524.25);
    a = day + 1 + x - (long)(x / 4.0);
    b = a + 1524;
    c = (long)(((double)b -122.1) /365.25);
    d = (long)(365.25 * c);
    e = (long)((double)(b - d) / 30.6001);
    *dd = b - d - (long)(30.6001 * e);
    if (e <= 13) {
        *mm = e - 1;
        *yy = c - 4716;
    }
    else {
        *mm = e - 13;
        *yy = c - 4715;
    }
}

void OLE::double2time(double v, int *y, int *m, int *d, int *hh, int *mm, int *ss){
    long day;
    double t;

    day = (long)v;
    t = v - day;
    jd2civil(2415019 + day, y, m, d);

    d2time(t, hh, mm, ss);
}

void OLE::d2time(double v, int *hh, int *mm, int *ss){
    double d_hh, d_mm, d_ss;
    int    i_hh, i_mm, i_ss;

    double d = v * 86400.0;

    d_hh = d / 3600.0;
    i_hh = (int)d_hh;

    d = d - i_hh * 3600.0;

    d_mm = d / 60.0;
    i_mm = (int)d_mm;

    d = d - i_mm * 60.0;

    d_ss = d * 10.0 + 5;
    
    i_ss = (int)d_ss / 10;

    if(i_ss == 60) {
        i_mm += 1;
        i_ss = 0;
    }

    if (i_mm == 60) {
        i_hh += 1;
        i_mm = 0;
    }
    if (i_hh == 24) {
        i_hh = 0;
    }
    
    *hh = i_hh;
    *mm = i_mm;
    *ss = i_ss;
}

intptr_t OLE::Handles::add(IDispatch *d,bool release){
	struct funcforfindif{bool operator()(std::pair<IDispatch*,bool> a){return a.first == NULL;}};
	std::vector<std::pair<IDispatch*,bool> >::iterator ite = std::find_if(m_disps.begin(),m_disps.end(),funcforfindif());
	if(ite != m_disps.end()){
		ite->first = d;ite->second = release;
		return (ite - m_disps.begin()) + 1;
	}else{
		std::pair<IDispatch*,bool> tmp;
		tmp.first = d;tmp.second = release;
		m_disps.push_back(tmp);
		return m_disps.size();
	}
}

IDispatch *OLE::Handles::operator[](int n){
	if(n <= 0 || n > static_cast<int>(m_disps.size()))return NULL;
	else return m_disps[n-1].first;
}

void OLE::Handles::clear(){
	for(std::vector<std::pair<IDispatch*,bool> >::iterator ite = m_disps.begin() ; ite != m_disps.end() ; ++ite){
		if(ite->second)ite->first->Release();
	}
	m_disps.clear();
}