#ifndef __COMPILE_H__
#define __COMPILE_H__

/** RpCp̉ʂ
*/

#include <editor/editwnd.h>
#include <window/basewnd.h>
#include <system/Process.h>
#include <ablib/Lock.h>
#include <boost/thread.hpp>
#include <boost/shared_ptr.hpp>
#include <vector>

class Compile : public ablib::window::BaseWnd,ablib::Lock{
public:
	Compile();
	~Compile();
	int Set(const TCHAR *set,const TCHAR *val);
	int SetEnv(const TCHAR *name,const TCHAR *val){m_envs.push_back(ENV(name,val));return TRUE;}
	int ClearEnv(){m_envs.clear();return TRUE;}
	int SetSize(int width,int height){m_width = width;m_height = height;return TRUE;}
	int SetFont(const TCHAR *font,int size){
		ablib::window::edit::EditWnd::EditConfig c = m_edit.GetConfig();
		c.fontname = font;c.fontsize = size;m_edit.SetConfig(c);return TRUE;
	}
	int Start(HWND hwnd,int x,int y,const TCHAR *command,const TCHAR *dir);
	int AddHilight(const TCHAR *keyword,int opt,int pos,COLORREF color,COLORREF back){
		m_hilight.push_back(Hilight(keyword,opt,pos,color,back));
		return TRUE;
	}
	int ClearHilight(){m_hilight.clear();return TRUE;}
	int SetColor(COLORREF color,COLORREF back){
		ablib::window::edit::EditWnd::EditConfig c = m_edit.GetConfig();
		c.NormalColor.normal = ablib::window::rgb(color);
		c.NormalColor.eof = c.NormalColor.Return = 
			c.NormalBackColor.eof = c.NormalBackColor.Return = c.NormalBackColor.normal = 
			ablib::window::rgb(back);
		m_edit.SetConfig(c);
		return TRUE;
	}
	int GetExitCode(){return m_process.GetExitCode();}
	void SetCode(int code){m_charcode = code;}

	int Open(HWND hwnd,int x,int y);
	int Close();

private:
	int m_charcode;
	template<typename T> struct CharBuf{
		CharBuf() : buf(NULL),size(0){}
		~CharBuf(){delete[] buf;}
		void Alloc(unsigned int s){
			if(s > size || buf == NULL){
				delete[] buf;
				buf = new T[2*s];
				size = 2*s;
			}
		}
		T *buf;
		unsigned int size;
	};

	void TransCode(CharBuf<char> &str,int length,ablib::string &s);

	struct ENV{
		ENV(){};
		ENV(const TCHAR *n,const TCHAR *v) : name(n),value(v){};
		ablib::string name;
		ablib::string value;
	};
	struct COMPILESET{
		ablib::string log;
		ablib::string pos;
		ablib::string title;
	}m_set;
	struct Hilight{
		Hilight(const TCHAR *key,int opt,int pos,COLORREF col,COLORREF bac)
			: keyword(key),option(opt),position(pos),
			color(ablib::window::rgb(col)),back(ablib::window::rgb(bac)){}
		ablib::string keyword;
		int option,position;
		ablib::window::rgb color,back;
	};


	class Edit : public ablib::window::edit::EditWnd{
	public:
		Edit(HINSTANCE appint) : EditWnd(appint){m_keybuf.reserve(20);}
		void Set(Compile *p){m_parent = p;}
	private:
		bool OnChar(ablib::window::Window *wnd,TCHAR ch,int repeat);
		bool OnKey(ablib::window::Window *wnd,bool isDown,UINT vk,int repeat){
			if(vk == VK_BACK)return false;
			else return EditWnd::OnKey(wnd,isDown,vk,repeat);
		}
		bool OnRButtonDown(ablib::window::Window *wnd,ablib::window::Mouse *mos,bool isDClick);
		ablib::string m_keybuf;
		Compile *m_parent;
	}m_edit;
	
	std::vector<Hilight> m_hilight;
	ablib::system::Process m_process;
	std::vector<ENV> m_envs;
	FILE *m_logfp;
	boost::shared_ptr<boost::thread> m_thread;
	HANDLE m_event;
	bool m_openwindow,m_read;
	void MsgLoop(HWND hwnd,int x,int y);
	int StartProcess(HWND parent,const TCHAR *command,const TCHAR *dir);

	int m_width,m_height;

	bool OnCreate(ablib::window::Window *wnd);
	bool OnSize(ablib::window::Window *wnd,int sw,int width,int height);
	bool OnSetFocus(ablib::window::Window *wnd,ablib::window::Window *old);
	bool OnClose(ablib::window::Window *wnd);
	bool OnDestroy(ablib::window::Window *wnd);

	friend Edit;

#ifndef NDEBUG
	void Dump(const CharBuf<char> &b);
#endif
};

#endif //__COMPILE_H__

