#include "stdafx.h"
#include <ablib/lexical_cast.h>
#include "OLE.h"

extern OLE *g_OLE;

OLE::OLE(){
	HRESULT hr = ::CoInitialize(NULL);
	::VariantInit(&m_result);
}

OLE::~OLE(){
	Clear();
	::VariantClear(&m_result);
	::CoUninitialize();
}

void OLE::Clear(){
	for(std::vector<IDispatch*>::size_type i = 0 ; i < m_disps.size() ; ++i){
		if(m_disps[i]){
			m_disps[i]->Release();
			m_disps[i] = NULL;
		}
	}
	m_disps.clear();
}

OLECHAR *OLE::GetOleChar(const char *str){
	if(str == NULL)return NULL;
	int len = ::lstrlen(str);
	int size = ::MultiByteToWideChar(CP_ACP,0,str,len,NULL,0);
	LPWSTR buf = ::SysAllocStringLen(NULL,size);
	::MultiByteToWideChar(CP_ACP,0,str,len,buf,size);
	return buf;
}

int OLE::Initialize(const TCHAR *obj){
	CLSID cls;
	#ifdef UNICODE
	HRESULT hr = ::CLSIDFromProgID(obj,&cls);
	if(FAILED(hr)){
		hr = ::CLSIDFromString(buf,&cls);
	}
	#else
	OLECHAR *buf = GetOleChar(obj);
	HRESULT hr = ::CLSIDFromProgID(buf,&cls);
	if(FAILED(hr)){
		hr = ::CLSIDFromString(buf,&cls);
	}
	::SysFreeString(buf);
	#endif
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("͒mȂOLET[ołD"));
		return NULL;
	}
	
	IDispatch *disp = NULL;
	hr = ::CoCreateInstance(cls,NULL,CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER,
	IID_IDispatch,reinterpret_cast<void**>(&disp));
	//	hr = ::CoCreateInstance(cls,NULL,CLSCTX_ALL,IID_IDispatch,(void**)&disp);
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("̏ɎsD"));
		return NULL;
	}
	m_disps.push_back(disp);
	//	disp->Release();
	return reinterpret_cast<int>(disp);
}

int OLE::Connect(const TCHAR *obj){
	CLSID cls;
	#ifdef UNICODE
	HRESULT hr = ::CLSIDFromProgID(obj,&cls);
	if(FAILED(hr)){
		hr = ::CLSIDFromString(buf,&cls);
	}
	#else
	OLECHAR *buf = GetOleChar(obj);
	HRESULT hr = ::CLSIDFromProgID(buf,&cls);
	if(FAILED(hr)){
		hr = ::CLSIDFromString(buf,&cls);
	}
	::SysFreeString(buf);
	#endif
	if(FAILED(hr))return Bind(obj);
	
	IUnknown *unknown;
	hr = ::GetActiveObject(cls,0,&unknown);
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("͓Ă܂D"));
		return NULL;
	}
	IDispatch *disp = NULL;
	hr = unknown->QueryInterface(IID_IDispatch,reinterpret_cast<void**>(&disp));
	unknown->Release();
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("Ƃ̐ڑɎsD"));
		return NULL;
	}
	m_disps.push_back(disp);
	return reinterpret_cast<int>(disp);
}

int OLE::Bind(const TCHAR *obj){
	IBindCtx *bindctx;
	HRESULT hr = ::CreateBindCtx(0,&bindctx);
	if(FAILED(hr)){
		SetmacrodllLastError(_T("CreateBindCtxɎs"));
		return NULL;
	}
	
	ULONG eaten;
	IMoniker *moniker;
#ifdef UNICODE
	hr = ::MkParseDisplayName(bindctx,obj,&eaten,&moniker);
#else
	OLECHAR *buf = GetOleChar(obj);
	hr = ::MkParseDisplayName(bindctx,buf,&eaten,&moniker);
	::SysFreeString(buf);
#endif
	if(FAILED(hr)){
		bindctx->Release();
		SetmacrodllLastError(ablib::string(obj) + _T("ɑ΂CMkParseDisplayNameŎsD"));
		return NULL;
	}
	IDispatch *disp;
	hr = moniker->BindToObject(bindctx,NULL,IID_IDispatch,reinterpret_cast<void**>(&disp));
	moniker->Release();
	bindctx->Release();
	if(FAILED(hr)){
		SetmacrodllLastError(ablib::string(obj) + _T("bindsD"));
		return NULL;
	}
	m_disps.push_back(disp);
	return reinterpret_cast<int>(disp);
}


int OLE::SetPropertyStr(int handle,const TCHAR *prop,const TCHAR *val){
	IDispatch *disp = reinterpret_cast<IDispatch*>(handle);
	VARIANT var;
	::VariantInit(&var);
	var.vt = VT_BSTR;
#ifdef UNICODE
	var.bstrVal = val;
	HRESULT hr = Invoke(handle,prop,1,var,NULL,DISPATCH_PROPERTYPUT);
#else
	var.bstrVal = GetOleChar(val);
	HRESULT hr = Invoke(handle,prop,1,&var,NULL,DISPATCH_PROPERTYPUT);
	::SysFreeString(var.bstrVal);
#endif
	::VariantClear(&var);
	return FAILED(hr) ? FALSE : TRUE;
}

int OLE::SetPropertyInt(int handle,const TCHAR *prop,int val){
	IDispatch *disp = reinterpret_cast<IDispatch*>(handle);
	VARIANT var;
	::VariantInit(&var);
	var.vt = VT_I4;
	var.intVal = val;
	HRESULT hr = Invoke(handle,prop,1,&var,NULL,DISPATCH_PROPERTYPUT);
	::VariantClear(&var);
	return FAILED(hr) ? FALSE : TRUE;
}


HRESULT OLE::Invoke(int handle, const TCHAR *func, int argc, VARIANTARG *argv, VARIANT *result, USHORT invoke_type){
	if(handle == 0){
		SetmacrodllLastError(_T("IuWFNgłD"));
		return FALSE;
	}
	ablib::string err;
	if(invoke_type == DISPATCH_METHOD)err = ablib::string(func) + _T("̎sɎs");
	else if(invoke_type == DISPATCH_PROPERTYGET)err = ablib::string(_T("vpeB")) + func + _T("擾ł܂ł");
	else if(invoke_type == DISPATCH_PROPERTYPUT)err = ablib::string(_T("vpeB")) + func + _T("ݒł܂ł");
	IDispatch *disp = reinterpret_cast<IDispatch*>(handle);
	DISPID dispid;
#ifdef UNICODE
	HRESULT hr = disp->GetIDsOfNames(IID_NULL,prop,1,LOCALE_USER_DEFAULT,&dispid);
#else
	OLECHAR *buf = GetOleChar(func);
	HRESULT hr = disp->GetIDsOfNames(IID_NULL,&buf,1,LOCALE_USER_DEFAULT,&dispid);
	::SysFreeString(buf);
#endif
	if(FAILED(hr)){
		SetmacrodllLastError(err + _T("iGetIDsOfNamesŎsjD"));
		return 0;
	}
	
	DISPPARAMS dispParams;
	dispParams.rgvarg = argv;
	dispParams.rgdispidNamedArgs = NULL;
	dispParams.cArgs = argc;
	dispParams.cNamedArgs = 0;
	
	DISPID dispIDNamedArgs[1] = {DISPID_PROPERTYPUT};
	if(invoke_type == DISPATCH_PROPERTYPUT){
		dispParams.cNamedArgs = 1;
		dispParams.rgdispidNamedArgs = dispIDNamedArgs;
	}
	
	EXCEPINFO excepinfo;
	::ZeroMemory(&excepinfo,sizeof(EXCEPINFO));
	UINT argerr = 0;
	hr = disp->Invoke(dispid,IID_NULL,LOCALE_USER_DEFAULT,invoke_type,&dispParams,result,&excepinfo,&argerr);
	if(FAILED(hr)){
		SetmacrodllLastError(err + _T("iInvokeŎsjD\n" + GetErrMsg(&excepinfo)));
		return 0;
	}
	return TRUE;
}

ablib::string OLE::GetErrMsg(EXCEPINFO *ei){
	if(ei->pfnDeferredFillIn)(ei->pfnDeferredFillIn)(ei);
	ablib::string rv;
	if(ei->bstrSource){
		rv += BSTR2String(ei->bstrSource);
	}else{
		rv += "ssource";
	}
	rv += "ɂāCR[h";
	if(ei->wCode == 0)rv += ablib::lexical_cast<ablib::string>(ei->scode);
	else rv += ablib::lexical_cast<ablib::string>(ei->wCode);
	rv += "ŃG[D";
	if(ei->bstrDescription)rv += "\n" + BSTR2String(ei->bstrDescription);
	return rv;
}


ablib::string OLE::BSTR2String(BSTR bstr){
#ifdef UNICODE
	return ablib::string(bstr);
#else
	ablib::string r;
	int len = ::lstrlenW(bstr);
	int size = ::WideCharToMultiByte(CP_ACP,0,bstr,len,NULL,0,NULL,NULL);
	char *s = new char[size + 1];
	::ZeroMemory(s,sizeof(char)*(size + 1));
	::WideCharToMultiByte(CP_ACP,0,bstr,len,s,size,NULL,NULL);
	r = s;
	delete[] s;
	return r.c_str();
#endif
}

int OLE::GetLastResultInt(){return m_result.intVal;}
int OLE::GetLastResultDisp(){return reinterpret_cast<int>(m_result.pdispVal);}
const TCHAR *OLE::GetLastResultStr(){
	static ablib::string r;
	r = Variant2Str(m_result);
	return r.c_str();
}

int OLE::GetLastResultBool(){return m_result.boolVal;}

#define GETARGS \
		argc = ::lstrlen(arglist);\
		if(argc > 0){\
			argv = new VARIANTARG[argc];\
			va_list vlists;\
			va_start(vlists,arglist);\
			for(int i = 0 ; i < argc ; ++i){\
				int j = argc - i - 1;\
				::VariantInit(argv + j);\
				switch(arglist[i]){\
				case 's':\
					argv[j].vt = VT_BSTR;\
					argv[j].bstrVal = OLE::GetOleChar(va_arg(vlists,PTCHAR));\
					break;\
				case 'd':\
					argv[j].vt = VT_DISPATCH;\
					argv[j].pdispVal = reinterpret_cast<IDispatch*>(va_arg(vlists,int));\
					break;\
				case 'i':\
				default:\
					argv[j].vt = VT_I4;\
					argv[j].intVal = va_arg(vlists,int);\
					break;\
				}\
			}\
		}\
	
#define RELEASEARGS \
	for(int i = 0 ; i < argc ; ++i){\
		if(argv[i].vt == VT_BSTR)::SysFreeString(argv[i].bstrVal);\
		::VariantClear(argv + i);\
	}\
	delete[] argv;\


EXPORT int OLE_INVOKE(int handle,const TCHAR *func,const TCHAR *arglist,...){
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;

	if(arglist){
		GETARGS
	}
	
	HRESULT rv = g_OLE->Invoke_Sub(handle,func,argc,argv);

	RELEASEARGS
	
	return rv;
}

EXPORT int OLE_GETPROPERTYDISP(int handle,const TCHAR *prop,const TCHAR *arglist,...){
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;
	if(arglist){
		GETARGS
	}
	VARIANT var;
	::VariantInit(&var);
	HRESULT hr = g_OLE->GetProperty(handle,prop,&var,argc,argv);
	if(var.vt != VT_DISPATCH){
		SetmacrodllLastError(ablib::string(_T("vpeB")) + prop + _T("̌^VT_IDISPATCHł͂܂łD"));
	}
	IDispatch *rv = ((!FAILED(hr) && var.vt == VT_DISPATCH) ? var.pdispVal : NULL);
	::VariantClear(&var);
	RELEASEARGS
	return reinterpret_cast<int>(rv);
}

EXPORT int OLE_GETPROPERTYINT(int handle,const TCHAR *prop,const TCHAR *arglist,...){
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;
	if(arglist){
		GETARGS
	}
	VARIANT var;
	::VariantInit(&var);
	HRESULT hr = g_OLE->GetProperty(handle,prop,&var,argc,argv);
	int rv = var.intVal;
	::VariantClear(&var);
	RELEASEARGS
	return rv;
}

EXPORT const TCHAR *OLE_GETPROPERTYSTR(int handle, const TCHAR *prop,const TCHAR *arglist,...){
	SetmacrodllLastError(_T(""));
	VARIANTARG *argv = NULL;
	int argc = 0;
	if(arglist){
		GETARGS
	}
	VARIANT var;
	::VariantInit(&var);
	HRESULT hr = g_OLE->GetProperty(handle,prop,&var,argc,argv);
	static ablib::string rv;
	if(!FAILED(hr)){
		rv = OLE::Variant2Str(var);
	}
	::VariantClear(&var);
	RELEASEARGS
	return rv.c_str();
}

HRESULT OLE::GetProperty(int handle, const TCHAR *prop, VARIANT *var, int argc, VARIANTARG *argv){
	HRESULT rv = Invoke(handle,prop,argc,argv,var,DISPATCH_PROPERTYGET);
	if(!FAILED(rv) && var->vt == VT_DISPATCH){
		var->pdispVal->AddRef();
		m_disps.push_back(var->pdispVal);
	}
	return rv;
}

ablib::string OLE::Variant2Str(VARIANT &var){
	ablib::string rv;
	if(var.vt == VT_BSTR){
		rv = OLE::BSTR2String(var.bstrVal);
	}else if(var.vt == VT_DATE){
		DATE date;
		if(V_ISBYREF(&var))date = *V_DATEREF(&var);
		else date = V_DATE(&var);
		int y, m, d, hh, mm, ss;
		char szTime[20];
		OLE::double2time(date, &y, &m, &d, &hh, &mm, &ss);
		sprintf(szTime,"%4.4d/%02.2d/%02.2d %02.2d:%02.2d:%02.2d",y, m, d, hh, mm, ss);
		rv = szTime;
	}else{
		VARIANT v;
		::VariantInit(&v);
		HRESULT hr = ::VariantChangeTypeEx(&v,&var,LOCALE_USER_DEFAULT,0,VT_BSTR);
		if(!FAILED(hr))rv = OLE::BSTR2String(v.bstrVal);
		else{
			::SetmacrodllLastError(_T("ʂ̕ւ̕ϊɎsD"));
		}
		::VariantClear(&v);
	}
	return rv;
}
// win32olẽRsy
void OLE::jd2civil(long day,int *yy, int *mm, int *dd){
    long x, a, b, c, d, e;
    x = (long)(((double)day - 1867216.25) / 36524.25);
    a = day + 1 + x - (long)(x / 4.0);
    b = a + 1524;
    c = (long)(((double)b -122.1) /365.25);
    d = (long)(365.25 * c);
    e = (long)((double)(b - d) / 30.6001);
    *dd = b - d - (long)(30.6001 * e);
    if (e <= 13) {
        *mm = e - 1;
        *yy = c - 4716;
    }
    else {
        *mm = e - 13;
        *yy = c - 4715;
    }
}

void OLE::double2time(double v, int *y, int *m, int *d, int *hh, int *mm, int *ss){
    long day;
    double t;

    day = (long)v;
    t = v - day;
    jd2civil(2415019 + day, y, m, d);

    d2time(t, hh, mm, ss);
}

void OLE::d2time(double v, int *hh, int *mm, int *ss){
    double d_hh, d_mm, d_ss;
    int    i_hh, i_mm, i_ss;

    double d = v * 86400.0;

    d_hh = d / 3600.0;
    i_hh = (int)d_hh;

    d = d - i_hh * 3600.0;

    d_mm = d / 60.0;
    i_mm = (int)d_mm;

    d = d - i_mm * 60.0;

    d_ss = d * 10.0 + 5;
    
    i_ss = (int)d_ss / 10;

    if(i_ss == 60) {
        i_mm += 1;
        i_ss = 0;
    }

    if (i_mm == 60) {
        i_hh += 1;
        i_mm = 0;
    }
    if (i_hh == 24) {
        i_hh = 0;
    }
    
    *hh = i_hh;
    *mm = i_mm;
    *ss = i_ss;
}

